/* Copyright 1996 Free Software Foundation, Inc.
   Contributed by Marcin Dalecki <dalecki@sub994.sub.uni-goettingen.de>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <errno.h>

#include "util.h"
#include "misc.h"

/*
 * Remove all trailing white spaces in place.
 * Return the new endaddress of the string.
 */
char *
strip_end (char *str)
{
  int len = strlen (str);
  str += len - 1;

  while (len > 0 && (isspace (*str) || *str == 26))
    {
      *str = '\0';
      --len;
      --str;
    }
  return str + 1;
}

/*
 * Remove a pointer to the first non whitespace char.
 */
char *
str_skip (const char *str)
{
  while (isspace (*str))
    str++;
  return (char *) str;
}

/*
 * Free an table of strings.
 */
void
tbstr_free (char *lst[], int n)
{
  while (n > 0)
    free (lst[--n]);
}

char *
stpcpy (char *dst, const char *src)
{
  strcpy (dst, src);

  return dst + strlen (dst);
}

/*
 * Remove the leading path from a filename.
 */
void
path_extrname (const char *path, char *name)
{
  const char *pt = strrchr (path, '/');
  pt = pt == NULL ? path : pt + 1;
  strcpy (name, pt);
}


/*
 * Strip the extension of a file name
 * Return a pointer to a static buffer
 */
const char *
stripo (const char *fname)
{
  const char *pt = fname;
  while ((pt = strchr (fname, '/')) != NULL)
    fname = pt + 1;
  if ((pt = strrchr (fname, '.'))
      && ((strcmp (pt, ".o") == 0) || (strcmp (pt, ".mod") == 0)))
    {
      char *leak = (char *) xmalloc (pt - fname + 1);
      strncpy (leak, fname, pt - fname);
      leak[pt - fname] = '\0';
      fname = leak;
    }
  return fname;
}

/*
 * Read a line of a configuration file and process continuation line.
 * Return buf, or NULL si eof
 * Blank at the end of line are always stripped.
 * Everything following comcar is a comment.  The line is cut before.
 * "noline will be updated if supplied and will contain the line number.
 */
char *
fgets_strip (char *buf, int sizebuf, FILE * fin, int *noline)
{
  int nocomment = 1;		/* No comments found ? */

  int contline = 0;
  char *debut = buf;
  char *ret = NULL;
  *buf = '\0';
  while (fgets (buf, sizebuf, fin) != NULL)
    {
      char *end = strip_end (buf);
      char *pt = strchr (buf, COMMENT_CHAR);
      if (pt != NULL)
	{
	  nocomment = 0;
	  *pt = '\0';
	  end = strip_end (buf);
	}
      if (noline != NULL)
	(*noline)++;
      ret = debut;
      if (contline)
	{
	  char *pt = str_skip (buf);
	  if (pt > buf + 1)
	    {
	      strcpy (buf + 1, pt);
	      buf[0] = ' ';
	      end -= (int) (pt - buf) - 1;
	    }
	  else if (pt == buf + 1)
	    {
	      buf[0] = ' ';
	    }
	}
      if (end > buf && *(end - 1) == LINECAT_CHAR)
	{
	  if (end == buf + 1 || *(end - 2) != LINECAT_CHAR)
	    {
	      /* Continuation demand */
	      contline = 1;
	      end--;
	      *end = '\0';
	      buf = end;
	    }
	  else
	    {
	      *(end - 1) = '\0';
	      break;
	    }
	}
      else
	{
	  break;
	}
    }
  return ret;
}

/*
 * Interface to the symbol loading facilities
 */
int debugmode = 0;

struct kernel_sym *ksymtab = NULL;
int nksyms;

void
load_kernel_symbols (void)
{
  nksyms = get_kernel_syms (NULL);

  if (nksyms < 0)
    {
      lprintf ("get_kernel_sys failed: Cannot find Kernel symbols!");
      exit (2);
    }

  if (ksymtab)
    free (ksymtab);

  if (nksyms)
    {
      ksymtab = (struct kernel_sym *) malloc (nksyms * sizeof *ksymtab);
      get_kernel_syms (ksymtab);
    }
}

/*
 * Error logging facilities.
 */

int log;			/* this is passed to insmod */
static int silent;

int errors;
const char *error_file;

void
error (const char *fmt,...)
{
  va_list args;

  if (silent)
    ;
  else if (log)
    {
      char buf[1024];
      int n;

      n = snprintf (buf, sizeof (buf), "%s: ",
		    error_file ? error_file : "insmod");
      va_start (args, fmt);
      vsnprintf (buf + n, sizeof (buf) - n, fmt, args);
      va_end (args);

      syslog (LOG_ERR, "%s", buf);
    }
  else
    {
      fprintf (stderr, "%s: ", error_file ? error_file : "insmod");
      va_start (args, fmt);
      vfprintf (stderr, fmt, args);
      va_end (args);
      putc ('\n', stderr);
    }

  errors++;
}

void
lprintf (const char *fmt,...)
{
  va_list args;

  if (silent)
    ;
  else if (log)
    {
      char buf[1024];
      va_start (args, fmt);
      vsnprintf (buf, sizeof (buf), fmt, args);
      va_end (args);
      syslog (LOG_INFO, "%s", buf);
    }
  else
    {
      va_start (args, fmt);
      vfprintf (stdout, fmt, args);
      va_end (args);
      putchar ('\n');
    }
}

void
setsyslog (const char *program)
{
  openlog (program, LOG_CONS, LOG_DAEMON);
  log = 1;
}
