/* Copyright 1997 Free Software Foundation, Inc.
   Contributed by Marcin Dalecki <dalecki@sub994.sub.uni-goettingen.de>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/param.h>
#include <getopt.h>
#include <sys/stat.h>

#include "util.h"
#include "misc.h"
#include "config.h"

/*
 * This is the actually modprobe specific part.
 *
 * The geenreall convention throughout this file is that functions are ALLWAYS
 * returning 0 on sccess and some different value otherwise.
 */

struct dep_node
  {
    struct dep_node *next;	/* modules */
    struct dep_node *deps;	/* dependences */
    char *name;
  };

/*
 * Command line flags
 */
static char verbose_flag = 0;
static int from_kerneld = 0;

/*
 * Dependencies of kernel resident and external modules are hold in this two
 * globals during loading.
 */
static struct dep_node *kernel_deps = NULL;
static struct dep_node *file_deps = NULL;

static void
verbose (const char *ctl,...)
{
  if (verbose_flag)
    {
      va_list list;
      va_start (list, ctl);
      vprintf (ctl, list);
      va_end (list);
      fflush (stdout);
    }
}

/*
 * Free the memmory associated with an dependency list
 */
static void
deps_free (struct dep_node *node)
{
  if (!node)
    return;

  free (node->name);
  deps_free (node->next);
  deps_free (node->deps);
  free (node);
}

static struct dep_node *
dep_lookup (struct dep_node *a, const char *name)
{
  if (!a)
    return NULL;

  if (strcmp (name, a->name) == 0)
    return a;

  return dep_lookup (a->next, name);
}

static struct dep_node *
dep_add (struct dep_node *node, const char *name)
{
  struct dep_node *tmp = (struct dep_node *)
  xmalloc (sizeof (struct dep_node));

  tmp->name = xstrdup (name);
  tmp->deps = NULL;
  tmp->next = node;

  return tmp;
}

/*
 *  Read the list of modules already loaded in the kernel
 */
static int
read_deps_kernel (void)
{
  int ret = 0;

  struct kernel_sym *ksym;
  int so_far = 0;

  deps_free (kernel_deps);
  kernel_deps = NULL;

  load_kernel_symbols ();

  for (ksym = ksymtab; so_far < nksyms; ++so_far, ksym++)
    if (ksym->name[0] == '#')
      {
	if (ksym->name[1])
	  kernel_deps = dep_add (kernel_deps, ksym->name + 1);
	else
	  break;
      }

  return ret;
}

/*
 * Read the dependancy file.
 */
static int
read_deps_file (const char *cfgfile)
{
  int line = 0;
  char *buf;
  char *cp;
  struct stat statb;
  FILE *fp;

  if (stat (cfgfile, &statb) == -1
      || (statb.st_mode & S_IFMT) != S_IFREG
      || !(fp = fopen (cfgfile, "r")))
    {
      lprintf ("Can't open dependancies file %s (%s)",
	       cfgfile, strerror (errno));
      return 1;
    }

  /*
   * OK: Now first read all of the file into an dynamically allocated buffer.
   */

  buf = (char *) xmalloc ((unsigned) (statb.st_size + 1));
  if (!fread (buf, sizeof (char), statb.st_size + 1, fp))
      lprintf ("Warning: empty dependency file!\n");


  buf[statb.st_size] = '\0';	/* be sure to NULL-terminate */
  fclose (fp);

  /*
   * And now parse the buffer.
   */
  cp = buf;

  /*
   * Go through the whole buffer and remove comments by blanking them out
   */

  while (*cp)
    if (*cp == '#')
      while (*cp && *cp != '\n')
	{
	  *cp = ' ';
	  ++cp;
	}
    else
      ++cp;

  cp = buf;

  while (*cp)
    {
      char *name;
      char *end;

      /*
       * Look for the module's name.
       */

      /* skip leading white spaces and newlines */
      while (*cp && isspace (*cp))
	{
	  if (*cp == '\n')
	    ++line;
	  ++cp;
	}
      if (!*cp)
	break;			/* done */

      name = cp;		/* here it will start */

      /* now search for the ':' */
      while (*cp && (*cp != ':'))
	{
	  if (*cp == '\n')
	    {
	      free (buf);
	      return 1;		/* parsing error */
	    }
	  ++cp;
	}
      if (!*cp)
	{
	  free (buf);
	  return 1;		/* parsing error */
	}

      end = cp;

      /* go back removig trailing white spaces */
      --cp;
      while ((cp > name) && isspace (*cp))
	--cp;
      if (cp == name)
	{
	  free (buf);
	  return 1;		/* parsing error */
	}
      cp[1] = '\0';

      /* name found */
      file_deps = dep_add (file_deps, name);

      cp = end + 1;

      /*
       * Now parse all the modules it's depending on
       */

      do
	{
	  /* remove leading white spaces */
	  while (*cp && (*cp != '\n') && isspace (*cp))
	    ++cp;
	  if (!*cp || *cp == '\n')
	    break;		/* no dependencies found */

	  /* name found */
	  name = cp;
	  while (*cp && !isspace (*cp))
	    ++cp;
	  if (cp == name)
	    {
	      free (buf);
	      return 1;		/* parse error empty non white */
	    }
	  *cp = '\0';
	  ++cp;

	  file_deps->deps = dep_add (file_deps->deps, name);
	}
      while (*cp && (*cp != '\n'));
    }

  free (buf);
  return 0;
}

static int
exec_rmmod_cmd (const char *mod)
{
  int ret = 0;
  char *ex;
  mod = stripo (mod);

  if ((ex = exec_cmd (EXEC_PRE_REMOVE, mod)) != NULL)
    if (system (ex) != 0)
      lprintf ("pre-remove %s failed\n", mod);

  if ((ex = exec_cmd (EXEC_REMOVE, mod)) != NULL)
    {
      if ((ret = system (ex)) != 0)
	lprintf ("remove %s failed\n", mod);
    }
  else if ((ret = delete_module (stripo (mod))) < 0)
    perror (stripo (mod));

  if (!ret && (ex = exec_cmd (EXEC_POST_REMOVE, mod)) != NULL)
    {
      if (system (ex) != 0)
	lprintf ("post-remove %s failed\n", mod);
    }

  return ret;
}

/*
 * Unload all sub-modules in reversese order.
 */
static int
rm_sub_mods (struct dep_node *nod)
{
  int ret = 0;
  if (nod != NULL)
    if (!(ret = rm_sub_mods (nod->next)))
      ret = exec_rmmod_cmd (nod->name);

  return ret;
}

/*
 * Return the options associated with a module.
 * Return NULL if there are none.
 */
static char *
any_options (const char *mod)
{
  const char *modname = stripo (mod);
  int len = strlen (modname);
  int i;

  for (i = 0; optlist[i]; ++i)
    if (strncmp (optlist[i], modname, len) == 0 && isspace (optlist[i][len]))
      return str_skip (&(optlist[i][len]));

  return NULL;
}

/*
 * Try to load a module and the sub-modules needed by it.
 */
static int
insmod (const char *mod, struct dep_node **newin_kernel, char *options[])
{
  int err = 0;
  struct dep_node *dep;
  struct dep_node *nod;
  char cmd[300];
  char *op;
  char *ex;

  if (!mod)
    return 0;

  if (dep_lookup (kernel_deps, stripo (mod)))
    return 0;			/* allready in kernel */

  if (dep_lookup (*newin_kernel, mod))
    return 0;			/* allready in kernel */

  if (!(nod = dep_lookup (file_deps, mod)))
    {
      lprintf ("No dependancy information for module %s", mod);
      return 1;
    }

  dep = nod->deps;
  while (dep && !err)
    {
      err = insmod (dep->name, newin_kernel, NULL);
      dep = dep->next;
    }

  if (err)
    {
      rm_sub_mods (nod->deps);
      read_deps_kernel ();
      return 1;
    }

  sprintf (cmd, "/sbin/insmod %s %s %s %s", from_kerneld ? "-k" : ""
	   ,log ? "-s" : ""
	   ,insmod_opt ? insmod_opt : ""
	   ,mod);

  if (options && options[1] &&
      strchr (options[1], '='))
    {
      int nopt;

      for (nopt = 1; options[nopt]; ++nopt)
	{
	  if (strchr (options[nopt], '='))
	    {
	      strcat (cmd, " ");
	      strcat (cmd, options[nopt]);
	    }
	  else
	    break;
	}
    }
  else
    {
      if ((op = any_options (mod)))
	{
	  strcat (cmd, " ");
	  strcat (cmd, op);
	}
      if (options && options[0] &&
	  (op = any_options (options[0])))
	{
	  strcat (cmd, " ");
	  strcat (cmd, op);
	}
    }
  verbose ("\r\t%s\n\t\t", cmd);
  /*
   * For now, seems like insmod is not reporting
   * success or failure correctly, therefore we read
   * /proc/modules each time
   */

  if ((ex = exec_cmd (EXEC_PRE_INSTALL, mod)) != NULL)
    if ((err = system (ex)) != 0)
      lprintf ("pre-install %s failed\n", mod);

  if (!err)
    {
      if ((ex = exec_cmd (EXEC_INSTALL, mod)) != NULL)
	err = system (ex);
      else
	err = system (cmd);
    }

  if (!err && (ex = exec_cmd (EXEC_POST_INSTALL, mod)) != NULL)
    if ((err = system (ex)) != 0)
      lprintf ("post-install %s failed\n", mod);

  if (err)
    {
      rm_sub_mods (nod->deps);
      read_deps_kernel ();
      return 1;
    }

  *newin_kernel = dep_add (*newin_kernel, mod);

  return 0;
}

static int unload (const char *mod);

/*
 * Unload all dependant-modules in reversese order.
 */
static int
rm_dep_mods (struct dep_node *nod)
{
  int ret = 0;

  if (nod != NULL)
    if (!(ret = rm_sub_mods (nod->next)))
      ret = unload (nod->name);

  return ret;
}

/*
 * Unload a module and whichever modules where requiered by this module.
 * We blindly remove everything in order, even if a module is needed
 * by another module. The idea is that this module will refuse to
 * unload.
 */
static int
unload (const char *mod)
{
  int ret = 0;
  char *tbpath[1000];
  int nbl;
  int m;

  if (!mod)
    return 0;

  /*
   * If there is no information about a module in
   * the dependancy file, we simply call /sbin/rmmod
   * on the module without further checking.
   */
  if (!(nbl = config_locate (mod, tbpath, NULL)))
    return exec_rmmod_cmd (mod);

  for (m = 0; m < nbl; m++)
    {
      char *path = tbpath[m];

      /*
       * By doing a dep_lookup and the kernel_deps we can see if this
       * module is actually in the kernel.
       */
      if (dep_lookup (kernel_deps, stripo (path)))
	{
	  struct dep_node *nod;

	  if (!(nod = dep_lookup (file_deps, path)))
	    lprintf ("No dependancy information for module %s", mod);
	  else
	    {
	      ret = exec_rmmod_cmd (path);
	      read_deps_kernel ();

	      rm_dep_mods (nod->deps);
	    }
	}
    }
  tbstr_free (tbpath, nbl);

  return ret;
}

/*
 * Load modules specified on the command line
 */
static int
load_from_list (char *list[], int nb, const char *type, int loadall)
{
  int ret = -1;
  int i;

  for (i = 0; i < nb; i++)
    {
      struct dep_node *kernel_deps = NULL;
      char *tbpath[1000];

      /*
       * We can pass option to module in the modprobe's command line.
       * It goes like this:
       * #/sbin/modprobe module opt1=value opt2=value [ othermodule ...]
       * An option is a keyword followed by an equal sign and a value.
       * No space are allowed in the sequence, unless it is quoted.
       *
       * The option list end at the end of the list or at the
       * first non-option argument (a module).
       */
      if (strchr (list[i], '=') == NULL)
	{
	  int nbl = config_locate (list[i], tbpath, type);
	  if (nbl == 0)
	    lprintf ("Can't locate module %s", list[i]);
	  else
	    {
	      int m;
	      for (m = 0; m < nbl; m++)
		{
		  if (insmod (tbpath[m], &kernel_deps, &list[i]) == 0)
		    {
		      ret = 0;
		      if (!loadall)	/* stop after fist sccess */
			break;
		    }
		}
	      tbstr_free (tbpath, nbl);
	    }
	}

      deps_free (kernel_deps);
      kernel_deps = NULL;

      if (ret == 0 && !loadall)
	break;
    }

  return ret;
}

/*
 * Print all available module matching "pattern" and of a certain type.
 * type may be NULL.
 */
static void
print_list (const char *pattern, const char *type)
{
  char *lst[1000];
  int nb = config_lstmod (pattern, type, lst, 1);
  int i;
  for (i = 0; i < nb; i++)
    printf ("%s\n", lst[i]);
  tbstr_free (lst, nb);
}

/*
 * Print usage information ans exit.
 */
static void
usage (void)
{
  printf ("Usage: modprobe [-a] [ -t TYPE ] MODULE_1 MODULE_2 ...\n"
	  "       modprobe -c\n"
	  "Load MODULE_1 MODULE_2, and modules needed by them.\n\n"
	  "  -a, --all                  load all modules\n"
	  "  -c, --show-config          show the current modules configuration and exit\n"
	  "  -d, --debug                run in debug mode\n"
	  "  -k, --kernel-daemon        only used by the kernel daemon\n"
	  "  -l, --list                 list currently installed modules\n"
	  "  -r, --remove               unload modules from the kernel\n"
  "  -s, --system-log           use the system logging for error reporting\n"
	  "  -t TYPE,\n"
    "  -type TYPE                 restrict actions to modules of the TYPE\n"
	  "      --help                 display this help and exit\n"
	  "  -v, --verbose              run in verbose mode\n"
	"  -V, --version              output version information and exit\n"
	  "\n"
    );
}

static void
nothing (const char *str)
{
  lprintf ("argument missing for %s\n"
	   "Plese specify at least one module or a wildcard like \\*.", str);
}


int
main (int argc, char *argv[])
{
  int ret = 0;
  int loadall = 0;		/* Load only one module out of a list */

  char *type = NULL;		/* Search in all path[] */
  int flag_remove = 0;
  int list = 0;

  int opt_tag;

  if (argc == 1)
    {
      usage ();
      return 1;
    }

  if (config_read (NULL))
    return 1;

  if (read_deps_file (config_getdepfile ())
      || read_deps_kernel ())
    return 1;

  /*
   * Yes we are using getopts!
   */
  while (1)
    {
      static struct option long_opts[] =
      {
	{"all", 0, 0, 'a'},
	{"show-conf", 0, 0, 'c'},
	{"debug", 0, 0, 'd'},
	{"kernel-daemon", 0, 0, 'k'},
	{"list", 0, 0, 'l'},
	{"remove", 0, 0, 'r'},
	{"system-log", 0, 0, 's'},
	{"type", 1, 0, 't'},
	{"verbose", 0, 0, 'v'},
	{"version", 0, 0, 'V'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}		/* Table end tag */
      };
      int opt_ind = 0;

      opt_tag = getopt_long (argc, argv, "acklrst:vV", long_opts,
			     &opt_ind);
      if (opt_tag == -1)
	break;

      switch (opt_tag)
	{
	case 'a':
	  loadall = 1;
	  break;

	case 'c':
	  if (argc != 2)
	    {			/* allow it only beeing used exclusive */
	      usage ();
	      exit (1);
	    }
	  config_show ();	/* show configuration */
	  exit (0);

	case 'd':
	  debugmode = 1;
	  break;

	case 'k':
	  from_kerneld = 1;	/* called by kernel daemon */
	  break;

	case 'l':
	  list = 1;		/* list modules of certain kind */
	  break;

	case 'r':
	  flag_remove = 1;
	  break;

	case 's':
	  setsyslog ("modprobe");	/* use the syslog for reporting */
	  break;

	case 't':
	  type = xstrdup (optarg);
	  break;

	case 'v':
	  verbose_flag = 1;	/* give terse informations during run */
	  break;

	case '?':
	case 'h':
	  usage ();
	  exit (opt_tag == 'h' ? 1 : 0);
	  break;

	case 'V':
	  puts ("modprobe (Linux modutils) " DEPMOD_RELEASE);
	  if (argc != 2)
	    putchar ('\n');
	  break;

	default:
	  abort ();
	}
    }

  if (ret == -1)
    return 1;

  /*
   * Skip all automatically processed options.
   */
  argc -= optind;
  argv += optind;

  /*
   * argv now points to the first non-option argument
   * argc is the remaining argument count
   */

  if (flag_remove)
    {
      if (argc > 0)
	for (; argc > 0 && ret == 0; ++argv, --argc)
	  ret = unload (*argv);
      else
	nothing ("remove");
    }
  else if (list)
    {
      if (argc > 0)
	for (; argc > 0 && ret == 0; ++argv, --argc)
	  print_list (*argv, type);
      else
	print_list ("*", type);
    }
  else
    {
      if (argc > 0)
	ret = load_from_list (argv, argc, type, loadall);
      else
	nothing ("load");
    }

  verbose ("\n");
  return ret;
}
